/* BBBloglines - a BB4W plugin to check the number of unread messages
   Copyright (C) 2004 kana <nicht AT s8 DOT xrea DOT con>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <BBApi-C.h>
#include <BBCmd.h>
#include <BBCmd_p.h>

#include "slhttp.h"


#define NUMBER_OF(array) (sizeof(array) / sizeof((array)[0]))




/* Constants */

static const char BBB_NAME[] = "BBBloglines";
static const char BBB_VERSION[] = "0.0.0";
static const char BBB_AUTHOR[] = "kana";
static const char BBB_RELEASE[] = "2004-10-16";
static const char BBB_LINK[] = "http://nicht.s8.xrea.com/2004/10/BBBloglines";
static const char BBB_EMAIL[] = "nicht AT s8 DOT xrea DOT con";

static const char BBB_CLASS_NAME[] = "BBBloglines";
static const char BBB_RC_NAME[] = "BBBloglines.rc";
static const UINT BBB_MESSAGES[] = {BB_BROADCAST, BB_RECONFIGURE, 0};

#define IDT_CHECK 1




/* Global variables */

	/**
	 * These are set in beginPlugin().
	 * And these will never be changed until endPlugin().
	 */
static HWND G_PluginWnd = NULL;
static HMODULE G_PluginModule = NULL;


	/* These are set only in read_settings(). */
#define G_CheckInterval_DEFAULT 10
static int G_CheckInterval = G_CheckInterval_DEFAULT;

static BBCmdCommand* G_CmdOnNewUnreadItem = NULL;
static BBCmdCommand* G_CmdOnUnreadItem = NULL;
static BBCmdCommand* G_CmdOnNoUnreadItem = NULL;
static BBCmdCommand* G_CmdOnError = NULL;

#define G_UnreadItems_DEFAULT 0
static int G_UnreadItems = G_UnreadItems_DEFAULT;

#define G_NotifierAPI_DEFAULT \
        "http://rpc.bloglines.com/update?user=<USER>&ver=1"
static char G_NotifierAPI[256] = G_NotifierAPI_DEFAULT;

#define G_ArgUser_DEFAULT ""
static char G_ArgUser[256] = G_ArgUser_DEFAULT;

	/* This is set only in check(). */
static int G_PreviousUnreadItems;

static const char G_ErrorMessage_DEFAULT[] = "No error";
static const char* G_ErrorMessage = G_ErrorMessage_DEFAULT;








/* Misc. functions */

static void
VSNPRINTF(char* buf, int size, const char* format, va_list va)
{
	if (buf == NULL || size <= 0) return;

	vsnprintf(buf, size, format, va);

	buf[size - 1] = '\0';
}


static void
SNPRINTF(char* buf, int size, const char* format, ...)
{
	va_list va;

	va_start(va, format);
	VSNPRINTF(buf, size, format, va);
	va_end(va);
}


/**
 * Replace `pattern' with `replacement' at once.
 * Return 0 for success or -1 for failure.
 *
 * xxxPATyyy
 *    |
 *    v
 * xxxREPLyyy
 */
int
strrepl(char* buf, int size, const char* pattern, const char* replacement)
{
	char* p;
	int buf_len;
	int pattern_len;
	int replacement_len;

	buf_len = strlen(buf);
	pattern_len = strlen(pattern);
	replacement_len = strlen(replacement);
	if (size < buf_len + replacement_len - pattern_len + 1)
		return -1;

	p = strstr(buf, pattern);

	memmove( p + replacement_len,
	         p + pattern_len,
	         sizeof(char) * (strlen(p + pattern_len) + 1) );
	memcpy(p, replacement, sizeof(char) * replacement_len);

	return 0;
}




static void
notice(const char* format, ...)
{
	char buf[80 * 25];
	va_list va;

	va_start(va, format);
	VSNPRINTF(buf, NUMBER_OF(buf), format, va);
	va_end(va);

	MessageBox( G_PluginWnd, buf, BBB_NAME,
	            MB_OK | MB_SETFOREGROUND | MB_TOPMOST);
}




static BOOL
is_2k_or_later(void)
{
	OSVERSIONINFO os;

	os.dwOSVersionInfoSize = sizeof(os);
	GetVersionEx(&os);

	return 5 <= os.dwMajorVersion;
}




static BOOL
get_module_dir(char* buf, int size)
{
	char* bs;

	if (!GetModuleFileName(G_PluginModule, buf, size))
		return FALSE;
	buf[size - 1] = '\0';

	bs = strrchr(buf, '\\');
	if (bs != NULL)
		bs[1] = '\0';

	return TRUE;
}


static BOOL
get_rc_path(char* buf, int size, const char* rc_name)
{
	char dir[MAX_PATH];

	if (!get_module_dir(dir, NUMBER_OF(dir)))
		return FALSE;

	if (size < strlen(dir) + strlen(rc_name) + 1)
		return FALSE;

	strcpy(buf, dir);
	strcat(buf, rc_name);

	return TRUE;
}




static BBCmdCommand*
read_script(LPCSTR rc, LPCSTR key, BBCmdCompiler* compiler)
{
	LPCSTR s;
	BBCmdCommand* cmd;
	int error_line;
	int error_column;
	const char* error_message;

	s = ReadString(rc, key, NULL);
	if (s == NULL) return NULL;

	cmd = BBCmdCompiler_CompileFromString(
		  compiler, s,
		  &error_line, &error_column, &error_message
	      );
	if (cmd == NULL) {
		notice( "In %s\n"
		        "\n"
		        "%s %s\n"
		        "\n"
		        "At column %d: %s",
		        BBB_RC_NAME,
			key, s,
			error_column, error_message );
	}

	return cmd;
}




/* `xbroam' command implementation */

typedef struct _BBCmdXBroamCommand {
	BBCmdCommand_HEAD

	char* xb_broam;
} BBCmdXBroamCommand;


/* forward */
static const BBCmdCommandDescriptor _BBCmdXBroamCommandDescriptor;
#define BBCmdXBroamCommandDescriptor (&_BBCmdXBroamCommandDescriptor)




static const char*
_BBCmdXBroamCommand_initialize( BBCmdCommand* _self,
                                  int ac, const char* const* av,
                                  const BBCmdCompiler* compiler )
{
	BBCmdXBroamCommand* self;
	const char* s;

	assert(_self->cmd_descriptor == BBCmdXBroamCommandDescriptor);
	self = (BBCmdXBroamCommand*)_self;

	self->xb_broam = NULL;

	if (ac - 1 != 1)
		return "Incorrect number of arguments";

	for (s = av[1]; *s != '\0'; s++) {
		if (*s == '%') {
			s++;

			switch (*s) {
			default:
				return "Invalid charcter after `%'";

			case '%':
			case 'n':
			case 'N':
			case 'e':
				break;
			}
		}
	}

	self->xb_broam = strdup(av[1]);
	if (self->xb_broam == NULL)
		return "Out of memory (xbroam)";

	return NULL;
}


static void
_BBCmdXBroamCommand_del(BBCmdCommand* _self)
{
	BBCmdXBroamCommand* self;

	assert(_self->cmd_descriptor == BBCmdXBroamCommandDescriptor);
	self = (BBCmdXBroamCommand*)_self;

	free(self->xb_broam);
}


static int
_BBCmdXBroamCommand_execute(const BBCmdCommand* _self)
{
	BBCmdXBroamCommand* self;
	char* broam;
	int len;
	int result;
	int i;

	assert(_self->cmd_descriptor == BBCmdXBroamCommandDescriptor);
	self = (BBCmdXBroamCommand*)_self;

	result = BBCmd_RC_FALSE;

	broam = strdup(self->xb_broam);
	if (broam == NULL) {
		notice("Out of memory (xbroam)");
		goto e_strdup;
	}

	len = strlen(broam);

	for (i = 0; broam[i] != '\0'; /*nop*/ ) {
		char tmp[64];
		int tmp_len;
		char* t;

		if (broam[i] != '%') {
			i++;
			continue;
		}

		switch (broam[i + 1]) {
		default:
			notice("Invalid conversion (xbroam)");
			goto e_convert;

		case '%':
			tmp[0] = '%';
			tmp[1] = '\0';
			goto case_convert;
		case 'n':
			SNPRINTF( tmp, NUMBER_OF(tmp), "%d",
			          G_UnreadItems - G_PreviousUnreadItems );
			goto case_convert;
		case 'N':
			SNPRINTF(tmp, NUMBER_OF(tmp), "%d", G_UnreadItems);
			goto case_convert;
		case 'e':
			SNPRINTF(tmp, NUMBER_OF(tmp), "%s", G_ErrorMessage);
			goto case_convert;
		case_convert:
			tmp_len = strlen(tmp);

			t = realloc(broam, sizeof(char) + (len + tmp_len + 1));
			if (t == NULL) {
				notice("Out of memory (xbroam)");
				goto e_convert;
			}

			broam = t;
			memmove( broam + i + tmp_len,
			         broam + i + 2,
			         sizeof(char) * (len - (i + 2) + 1) );
			memcpy(broam + i, tmp, sizeof(char) * tmp_len);
			len += tmp_len - 2;
			i += tmp_len;
			break;
		}
	}

	SendMessage(GetBBWnd(), BB_BROADCAST, 0, (LPARAM)broam);

	result = BBCmd_RC_TRUE;
e_convert:
	free(broam);
e_strdup:
	return result;
}


static const BBCmdCommandDescriptor _BBCmdXBroamCommandDescriptor = {
	"xbroam",
	NULL,
	"(broam)\n"
		"Send a specified `broam'\n"
		"after applying the following conversions:\n"
		"\n"
		"%%    Converted a `%'.\n"
		"%n    The number of new unread items.\n"
		"%N    The number of unread items.\n"
		"%e    The description of an error.",
	sizeof(BBCmdXBroamCommand),
	_BBCmdXBroamCommand_initialize,
	_BBCmdXBroamCommand_del,
	_BBCmdXBroamCommand_execute
};








/* Main functions */

#define RC_KEY(s) "bbbloglines." s ":"
#define RC_KEY_CHECK_INTERVAL RC_KEY("check_interval")
#define RC_KEY_ON_NEW_UNREAD_ITEM RC_KEY("on_new_unread_item")
#define RC_KEY_ON_UNREAD_ITEM RC_KEY("on_unread_item")
#define RC_KEY_ON_NO_UNREAD_ITEM RC_KEY("on_no_unread_item")
#define RC_KEY_ON_ERROR RC_KEY("on_error")
#define RC_KEY_UNREAD_ITEMS RC_KEY("unread_items")
#define RC_KEY_NOTIFIER_API RC_KEY("notifier_api")
#define RC_KEY_ARG_USER RC_KEY("arg.user")
#define RC_KEY_ARG_VER RC_KEY("arg.ver")


static void
read_settings(void)
{
	char rc[MAX_PATH];
	LPCSTR s;
	BBCmdCompiler* compiler;

	if (!get_rc_path(rc, NUMBER_OF(rc), BBB_RC_NAME)) {
		notice("Failed to get the path to %s", BBB_RC_NAME);
		goto E_RC_PATH;
	}

	compiler = BBCmdCompiler_New();
	if (compiler == NULL) {
		notice("No memory; failed to create a new compiler");
		goto E_BBCMD_COMPILER;
	}
	if (!BBCmdCompiler_AddNewCommand( compiler,
	                                  BBCmdXBroamCommandDescriptor )) {
		goto E_BBCMD_ADD_COMMAND;
	}


	G_CheckInterval = ReadInt( rc, RC_KEY_CHECK_INTERVAL,
	                           G_CheckInterval_DEFAULT );
	if (G_CheckInterval <= 0 || G_CheckInterval * 60 * 1000 <= 0) {
		notice( "In %s\n"
		        "\n"
		        "%s %d\n"
		        "\n"
		        "The value must be a positive and not too big number.",
			BBB_RC_NAME,
		        RC_KEY_CHECK_INTERVAL,
		        G_CheckInterval );
		G_CheckInterval = G_CheckInterval_DEFAULT;
	}

	if (G_CmdOnNewUnreadItem != NULL)
		BBCmdCommand_Del(G_CmdOnNewUnreadItem);
	G_CmdOnNewUnreadItem
		= read_script(rc, RC_KEY_ON_NEW_UNREAD_ITEM, compiler);

	if (G_CmdOnUnreadItem != NULL)
		BBCmdCommand_Del(G_CmdOnUnreadItem);
	G_CmdOnUnreadItem
		= read_script(rc, RC_KEY_ON_UNREAD_ITEM, compiler);

	if (G_CmdOnNoUnreadItem != NULL)
		BBCmdCommand_Del(G_CmdOnNoUnreadItem);
	G_CmdOnNoUnreadItem
		= read_script(rc, RC_KEY_ON_NO_UNREAD_ITEM, compiler);

	if (G_CmdOnError != NULL)
		BBCmdCommand_Del(G_CmdOnError);
	G_CmdOnError = read_script(rc, RC_KEY_ON_ERROR, compiler);

	G_UnreadItems = ReadInt( rc, RC_KEY_UNREAD_ITEMS,
	                         G_UnreadItems_DEFAULT );

	s = ReadString(rc, RC_KEY_ARG_USER, G_ArgUser_DEFAULT);
	SNPRINTF(G_ArgUser, NUMBER_OF(G_ArgUser), "%s", s);
	if (strcmp(G_ArgUser, G_ArgUser_DEFAULT) == 0) {
		notice( "In %s\n"
		        "\n"
		        "%s\n"
		        "\n"
		        "You have to specify the email address\n"
		        "of your Bloglines account for this option.",
			BBB_RC_NAME,
		        RC_KEY_ARG_USER );
	}

	s = ReadString(rc, RC_KEY_NOTIFIER_API, G_NotifierAPI_DEFAULT);
	SNPRINTF(G_NotifierAPI, NUMBER_OF(G_NotifierAPI), "%s", s);
	strrepl(G_NotifierAPI, NUMBER_OF(G_NotifierAPI), "<USER>", G_ArgUser);


E_BBCMD_ADD_COMMAND:
	BBCmdCompiler_Del(compiler);
E_BBCMD_COMPILER:
E_RC_PATH:
	return;
}


static void
write_settings(void)
{
	char rc[MAX_PATH];

	if (!get_rc_path(rc, NUMBER_OF(rc), BBB_RC_NAME)) {
		notice("Failed to get the path to %s", BBB_RC_NAME);
		return;
	}

	WriteInt(rc, RC_KEY_UNREAD_ITEMS, G_UnreadItems);

	if (strcmp(G_ArgUser, G_ArgUser_DEFAULT) == 0)
		WriteString(rc, RC_KEY_ARG_USER, G_ArgUser);

	/**
	 * These values don't have to be written back.
	 *
	 * WriteInt(rc, RC_KEY_CHECK_INTERVAL, G_CheckInterval);
	 * WriteString(rc, RC_KEY_ON_NEW_ITEM, ...);
	 * WriteString(rc, RC_KEY_ON_NO_ITEM, ...);
	 * WriteString(rc, RC_KEY_ON_ERROR, ...);
	 * WriteString(rc, RC_KEY_NOTIFIER_API, G_NotifierAPI);
	 */
}




static void
check(void)
{
	char user_agent[ ( (NUMBER_OF(BBB_NAME) + 1 + NUMBER_OF(BBB_VERSION))
	                   + 1 + (1 + NUMBER_OF(BBB_LINK) + 1) )
	                 + 1 ];
	int result;
	int status_code;
	char* status_line;
	char* headers;
	char* message_body;
	int n;
	BBCmdCommand* cmd;

	G_PreviousUnreadItems = G_UnreadItems;

	SNPRINTF( user_agent, NUMBER_OF(user_agent),
	          "%s/%s (%s)",
	          BBB_NAME, BBB_VERSION, BBB_LINK );

	result = slhttp_get( G_NotifierAPI, user_agent,
	                     &status_code,
	                     &status_line, &headers, &message_body );

	cmd = G_CmdOnError;

	if (result != SLHTTP_E_SUCCESS) {
		switch (result) {
		default:
			G_ErrorMessage = "Unknown error";
			break;

		case SLHTTP_E_INVALID_SCHEME:
		case SLHTTP_E_INVALID_URI:
			G_ErrorMessage = "The URI of notifier API is wrong";
			break;

		case SLHTTP_E_NO_MEMORY:
			G_ErrorMessage = "No memory to check";
			break;

		case SLHTTP_E_GETHOSTBYNAME:
			G_ErrorMessage
			  = "Failed to look up the IP address of Bloglines";
			break;

		case SLHTTP_E_SOCKET:
		case SLHTTP_E_CONNECT:
			G_ErrorMessage = "Failed to connect to Bloglines";
			break;

		case SLHTTP_E_INVALID_HTTP_RESPONSE:
			G_ErrorMessage
			  = "Invalid HTTP response from Bloglines";
			break;
		}
		goto e_slhttp_get;
	}

	if (status_code != 200) {
		G_ErrorMessage = "Failed to get information";
		goto e_status_code;
	}

	if (sscanf(message_body, "|%d|%*[^|]|", &n) != 1) {
		G_ErrorMessage = "Invalid data format";
		goto e_format;
	}

	if (n < 0) {
		G_ErrorMessage = "Wrong email address";
		goto e_account;
	}

	G_ErrorMessage = G_ErrorMessage_DEFAULT;

	G_UnreadItems = n;
	if (0 < G_UnreadItems - G_PreviousUnreadItems)
		cmd = G_CmdOnNewUnreadItem;
	else if (0 < G_UnreadItems)
		cmd = G_CmdOnUnreadItem;
	else
		cmd = G_CmdOnNoUnreadItem;

e_account:
e_format:
e_status_code:
	free(status_line);
	free(headers);
	free(message_body);
e_slhttp_get:
	if (cmd != NULL)
		BBCmdCommand_Execute(cmd);
	return;
}




void
interpret_broam(const char* broam)
{
	if (strncmp(broam, "@BBBloglines.", 13) != 0)
		return;
	broam += 13;

	if (strcmp(broam, "read_settings") == 0) {
		read_settings();
	} else if (strcmp(broam, "check_immediately") == 0) {
		check();

		/* Reset timer */
		SetTimer( G_PluginWnd,
		          IDT_CHECK, G_CheckInterval * 60 * 1000, NULL );
	} else {
		notice("Unknown bro@m: `%s'", broam - 13);
	}
}




static LRESULT CALLBACK
BBB_WndProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg) {
	default:
		break;


	case BB_BROADCAST:
		interpret_broam((const char*)lp);
		break;

	case BB_RECONFIGURE:
		read_settings();
		break;


	case WM_TIMER:
		if (wp == IDT_CHECK)
			check();
		break;


	case WM_CREATE:
		SendMessage( GetBBWnd(), BB_REGISTERMESSAGE,
		             (WPARAM)hwnd, (LPARAM)BBB_MESSAGES );
		SetTimer(hwnd, IDT_CHECK, G_CheckInterval * 60 * 1000, NULL);

		check();
		break;

	case WM_DESTROY:
		KillTimer(hwnd, IDT_CHECK);
		SendMessage( GetBBWnd(), BB_UNREGISTERMESSAGE,
		             (WPARAM)hwnd, (LPARAM)BBB_MESSAGES );
		break;
	}

	return DefWindowProc(hwnd, msg, wp, lp);
}








/* Public functions */

DLL_EXPORT int
beginPlugin(HINSTANCE hplugininstance)
{
	WNDCLASS wc;

	G_PluginModule = hplugininstance;

	if ( BBCmd_IsCompatible(BBCmd_VERSION, BBCmd_ICH_KNOW_INTERNAL)
	     != BBCmd_IC_COMPATIBLE )
	{
		goto E_BBCMD_COMPATIBILITY;
	}

	/* Read settings */
	read_settings();

	/* Create the window of this plugin */
	ZeroMemory(&wc, sizeof(wc));
	wc.lpfnWndProc = BBB_WndProc;
	wc.hInstance = hplugininstance;
	wc.lpszClassName = BBB_CLASS_NAME;
	if (!RegisterClass(&wc)) {
		goto E_REGISTER_CLASS;
	}

	G_PluginWnd = CreateWindowEx( WS_EX_TOOLWINDOW,
	                              BBB_CLASS_NAME,
	                              NULL,
	                              0,
	                              0, 0, 0, 0,
	                              (is_2k_or_later() ? HWND_MESSAGE : NULL),
	                              NULL,
	                              hplugininstance,
	                              NULL );
	if (G_PluginWnd == NULL) {
		goto E_CREATE_WINDOW;
	}

	return 0;


E_CREATE_WINDOW:
	UnregisterClass(BBB_CLASS_NAME, hplugininstance);
E_REGISTER_CLASS:
E_BBCMD_COMPATIBILITY:
	return !0;
}




DLL_EXPORT void
endPlugin(HINSTANCE hplugininstance)
{
	write_settings();

	DestroyWindow(G_PluginWnd);
	UnregisterClass(BBB_CLASS_NAME, hplugininstance);

	if (G_CmdOnNewUnreadItem != NULL)
		BBCmdCommand_Del(G_CmdOnNewUnreadItem);
	if (G_CmdOnUnreadItem != NULL)
		BBCmdCommand_Del(G_CmdOnUnreadItem);
	if (G_CmdOnNoUnreadItem != NULL)
		BBCmdCommand_Del(G_CmdOnNoUnreadItem);
	if (G_CmdOnError != NULL)
		BBCmdCommand_Del(G_CmdOnError);
}




DLL_EXPORT LPCSTR
pluginInfo(int field)
{
	switch (field) {
	default: /* FALLTHRU */
	case PLUGIN_NAME: return BBB_NAME;
	case PLUGIN_VERSION: return BBB_VERSION;
	case PLUGIN_AUTHOR: return BBB_AUTHOR;
	case PLUGIN_RELEASE: return BBB_RELEASE;
	case PLUGIN_LINK: return BBB_LINK;
	case PLUGIN_EMAIL: return BBB_EMAIL;
	}
}




/* __END__ */
